package com.jh.boot.i18n;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;


/**
 * TODO comment.
 *
 * @author jh
 */
public class I18nUtils {

	/** The date formats. */
	private Map<String, SimpleDateFormat> dateFormats;

	/** The date formatters. */
	private Map<String, DateTimeFormatter> dateFormatters;

	/** The number formats. */
	private Map<NumberFormatKey, NumberFormat> numberFormats;
	
	/** The instance. */
	private static I18nUtils instance = new I18nUtils();
	
	private I18nContext defaultI18n = new I18nContext() {
		
		@Override
		public String getThousandDelimiter() {
			return ",";
		}
		
		@Override
		public String getDecimalDelimiter() {
			return ".";
		}
		
		@Override
		public String getDateFormat() {
			return "MM/DD/YYYY";
		}
		
		@Override
		public String getCurrencySymbol() {
			return "$";
		}
		
		@Override
		public String getCurrencyFormat() {
			return "left";
		}
		
		@Override
		public String getCurrency() {
			return "USD";
		}
	};
	
	/**
	 * Instantiates a new i 18 n utils.
	 */
	private I18nUtils() {
		dateFormats = new HashMap<>();
		numberFormats = new HashMap<>();
		dateFormatters = new HashMap<>();
	}

	
	/**
	 * Format number.
	 *
	 * @param number the number
	 * @param i18n the i 18 n
	 * @return the string
	 */
	public String formatNumber(Number number, I18nContext i18n) {
		return formatNumber(number, 3, i18n);
	}
	
	/**
	 * Format number.
	 *
	 * @param number the number
	 * @param fixed the fixed
	 * @param i18n the i 18 n
	 * @return the string
	 */
	public String formatNumber(Number number, int fixed, I18nContext i18n) {
		if (number == null) {
			return "";
		}
		
		if (i18n == null) {
			i18n = defaultI18n;
		}
		
		NumberFormatKey nfk = new NumberFormatKey(i18n.getDecimalDelimiter(), i18n.getThousandDelimiter(), fixed);
		NumberFormat nf = numberFormats.get(nfk);
		if (nf == null) {
			DecimalFormatSymbols otherSymbols = new DecimalFormatSymbols();
			char ds = i18n.getDecimalDelimiter() == null || i18n.getDecimalDelimiter().isEmpty() ? '.' : i18n.getDecimalDelimiter().charAt(0);
			char gs = i18n.getThousandDelimiter() == null || i18n.getThousandDelimiter().isEmpty() ? '.' : i18n.getThousandDelimiter().charAt(0);
			
			otherSymbols.setDecimalSeparator(ds);
			otherSymbols.setGroupingSeparator(gs);
			
			DecimalFormat df = new DecimalFormat("#", otherSymbols);
			
			df.setGroupingUsed(true);
			df.setGroupingSize(3);
			
			df.setMinimumIntegerDigits(1);
			
			df.setMinimumFractionDigits(fixed);
			df.setMaximumFractionDigits(fixed);
			nf = df;
			
			numberFormats.put(nfk, df);
		}
		
		
		return nf.format(number);
	}
	
	
    /**
     * Format money.
     *
     * @param number the number
     * @param currency the currency
     * @param i18n the i 18 n
     * @param fixed the fixed
     * @return the string
     */
    public String formatMoney(Number number, String currency, I18nContext i18n, int fixed) {
        if (number == null) {
            return "";
        }
		if (i18n == null) {
			i18n = defaultI18n;
		}
		if (currency == null) {
			currency = "";
		}

        if (!"right".equals(i18n.getCurrencyFormat())) {
            StringBuilder sb = new StringBuilder(currency);
            sb.append(" ").append(formatNumber(number, fixed, i18n));
            return sb.toString();
        }
        
        StringBuilder sb = new StringBuilder(formatNumber(number, fixed, i18n));
        sb.append(" ").append(currency);
        return sb.toString();
    }
    
    
    /**
     * Format money.
     *
     * @param number the number
     * @param currency the currency
     * @param i18n the i 18 n
     * @return the string
     */
    public String formatMoney(Number number, String currency, I18nContext i18n) {
        return formatMoney(number, currency, i18n, 2);
    }
    
    
	/**
	 * Format percent.
	 *
	 * @param number the number
	 * @param i18n the i 18 n
	 * @return the string
	 */
	public String formatPercent(Number number, I18nContext i18n) {
		if (number == null) {
			return "";
		}

		StringBuilder sb = new StringBuilder(formatNumber(number, i18n));
		sb.append(" %");
		return sb.toString();
	}
	
	
	/**
	 * Format date.
	 *
	 * @param date the date
	 * @param i18n the i 18 n
	 * @return the string
	 */
	public String formatDate(Date date, I18nContext i18n) {
		if (date == null) {
			return "";
		}
		if (i18n == null) {
			i18n = defaultI18n;
		}

		String format = Objects.toString(i18n.getDateFormat(), "MM/dd/yyyy");
		SimpleDateFormat sdf = dateFormats.get(format);
		if (sdf == null) {
			dateFormats.put(format, sdf = new SimpleDateFormat(format));
		}
		return sdf.format(date);
	}

	
	/**
	 * Format date.
	 *
	 * @param date the date
	 * @param i18n the i 18 n
	 * @return the string
	 */
	public String formatDate(LocalDate date, I18nContext i18n) {
		if (date == null) {
			return "";
		}
		if (i18n == null) {
			i18n = defaultI18n;
		}

		String format = Objects.toString(i18n.getDateFormat(), "MM/dd/yyyy");
		DateTimeFormatter dtf = dateFormatters.get(format);
		if (dtf == null) {
			dateFormatters.put(format, dtf = DateTimeFormatter.ofPattern(convertJSFormat(format)));
		}
		return date.format(dtf);
	}

	
	/**
	 * Convert JS format.
	 *
	 * @param format the format
	 * @return the string
	 */
	private String convertJSFormat(String format) {
		return format.replace("D", "d").replace("m", "M").replace("Y","y");
	}


    /**
     * Gets the single instance of I18nUtils.
     *
     * @return single instance of I18nUtils
     */
    public static I18nUtils getInstance() {
        return instance;
    }
    
    
    class NumberFormatKey {
    	
    	/** The thousand delimiter. */
    	private String thousandDelimiter;
    	
    	/** The decimal delimiter. */
    	private String decimalDelimiter;
    	
    	/** The fixed. */
    	private int fixed;
    	
    	/**
    	 * Instantiates a new number format key.
    	 *
    	 * @param decimalDelimiter the decimal delimiter
    	 * @param thousandDelimiter the thousand delimiter
    	 * @param fixed the fixed
    	 */
    	public NumberFormatKey(String decimalDelimiter, String thousandDelimiter, int fixed) {
    		super();
    		this.thousandDelimiter = thousandDelimiter;
    		this.decimalDelimiter = decimalDelimiter;
    		this.fixed = fixed;
    	}

    	/**
    	 * @return
    	 * @see java.lang.Object#hashCode()
    	 */
    	@Override
    	public int hashCode() {
    		final int prime = 31;
    		int result = 1;
    		result = prime * result + ((decimalDelimiter == null) ? 0 : decimalDelimiter.hashCode());
    		result = prime * result + fixed;
    		result = prime * result + ((thousandDelimiter == null) ? 0 : thousandDelimiter.hashCode());
    		return result;
    	}

    	/**
    	 * @param obj
    	 * @return
    	 * @see java.lang.Object#equals(java.lang.Object)
    	 */
    	@Override
    	public boolean equals(Object obj) {
    		if (this == obj)
    			return true;
    		if (obj == null)
    			return false;
    		if (getClass() != obj.getClass())
    			return false;
    		NumberFormatKey other = (NumberFormatKey) obj;
    		if (decimalDelimiter == null) {
    			if (other.decimalDelimiter != null)
    				return false;
    		} else if (!decimalDelimiter.equals(other.decimalDelimiter))
    			return false;
    		if (fixed != other.fixed)
    			return false;
    		if (thousandDelimiter == null) {
    			if (other.thousandDelimiter != null)
    				return false;
    		} else if (!thousandDelimiter.equals(other.thousandDelimiter))
    			return false;
    		return true;
    	}
    	
    	
    }
    
    
}
