package com.jh.boot.template.handlebars;

import java.io.IOException;
import java.io.Writer;
import java.text.MessageFormat;
import java.time.LocalDate;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import javax.annotation.PostConstruct;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import com.github.jknack.handlebars.Context;
import com.github.jknack.handlebars.Handlebars;
import com.github.jknack.handlebars.Helper;
import com.github.jknack.handlebars.Options;
import com.github.jknack.handlebars.Template;
import com.github.jknack.handlebars.context.JavaBeanValueResolver;
import com.github.jknack.handlebars.context.MapValueResolver;
import com.jh.boot.i18n.I18nContext;
import com.jh.boot.i18n.I18nUtils;
import com.jh.boot.template.TemplateEngine;
import com.jh.boot.template.TemplateLoader;



/**
 * 
 * 
 * {{#if author}}
 * 	<h1>{{firstName}} {{lastName}}</h1>
 * {{else}}
 *	<h1>Unknown Author</h1>
 * {{/if}}
 * 
 * 
 * {{#each array}}
  {{@index}}: {{this}}
	{{/each}}
 *
 *
 *{{#each array as |value key|}}
  {{#each child as |childValue childKey|}}
    {{key}} - {{childKey}}. {{childValue}}
  {{/each}}
{{/each}}
 * 
 * {{#with author}}
  <p>{{name}}</p>
{{else}}
  <p class="empty">No content</p>
{{/with}}

  {{#with author as |myAuthor|}}
  <h2>By {{myAuthor.firstName}} {{myAuthor.lastName}}</h2>
  {{/with}}
 * 
 * Nested handlebars paths can also include ../ segments, which evaluate their paths against a parent context. 
 * 
 * 
 * @author Jan Hrabal
 */
public class HandlebarsTemplateEngine implements TemplateEngine {

	private boolean useCache = false;

	private Map<String, Template> templateCache;

	private Handlebars handlebars;
	
	private TemplateLoader templateLoader;
	
	
	/**
	 * Initializes instance.
	 */
	@PostConstruct
	public void init() {
		templateCache = new HashMap<>();
		handlebars = new Handlebars();
		handlebars.setParentScopeResolution(true);
		handlebars.setPreEvaluatePartialBlocks(false);

		//conditional helpers
		handlebars.registerHelper("eq", new Helper<Object>() {
			public CharSequence apply(Object a, Options options) throws IOException {
				Object b = options.param(0, null);
				boolean result = Objects.equals(a, b);
				if (!result) {
					result = Objects.equals(Objects.toString(a, null), Objects.toString(b, null));
				}
			    return result ? options.fn() : options.inverse();
			} 
		});
		
		I18nUtils i18nUtils = I18nUtils.getInstance();

		handlebars.registerHelper("formatNumber", new Helper<Number>() {
			public CharSequence apply(Number num, Options options) throws IOException {
				I18nContext i18nCtx = (I18nContext) options.param(0);
				return i18nUtils.formatNumber(num, i18nCtx);
			}
		});
		handlebars.registerHelper("formatPercent", new Helper<Number>() {
			public CharSequence apply(Number num, Options options) throws IOException {
				I18nContext i18nCtx = (I18nContext) options.param(0);
				return i18nUtils.formatPercent(num, i18nCtx);
			}
		});
		handlebars.registerHelper("formatMoney", new Helper<Number>() {
			public CharSequence apply(Number num, Options options) throws IOException {
				String currency = (String) options.param(0);
				I18nContext i18nCtx = (I18nContext) options.param(1);
				return i18nUtils.formatMoney(num, currency, i18nCtx);
			}
		});
		handlebars.registerHelper("formatDate", new Helper<LocalDate>() {
			public CharSequence apply(LocalDate date, Options options) throws IOException {
				I18nContext i18nCtx = (I18nContext) options.param(0);
				return i18nUtils.formatDate(date, i18nCtx);
			}
		});

		handlebars.registerHelper("formatLabel", new Helper<String>() {
			public CharSequence apply(String key, Options options) throws IOException {

				if (options.params == null || options.params.length < 1) {
					return "Wrong formatLabel syntax for key " + key + ". It should be formatLabel key labels ...params";
				}
				
				Map<String, Object> labels = (Map<String, Object>) options.params[0];
				if (labels == null) {
					return "!!" + key;
				}
				
				String label = (String) labels.get(key);
				if (label == null) {
					return "!!" + key;
				}
				
				Object[] messageParams = options.params.length > 1 ? Arrays.copyOfRange(options.params, 1, options.params.length) : new Object[0];
				return MessageFormat.format(label, messageParams);
			}
		});
	
	}
	
	
	@Override
	public void evaluate(Writer writer, String templateCode, Map<String, Object> data) {
	    String name = "template://" + templateCode;

	    Template template = useCache ? templateCache.get(name) : null;

		if (template == null) {
			com.jh.boot.template.Template t = templateLoader.loadTemplate(templateCode);
			if (t == null) {
//				throw new RuntimeException(String.format("Template %s not found", templateCode));
				return; //TODO log warning
			}
			try {
				template = handlebars.compileInline(t.getTemplate());
			} catch (IOException e) {
				throw new RuntimeException("Cannot evaluate template", e);
			}
			if (useCache) {
				templateCache.put(name, template);
			}
		}
		
		evaluateTemplate(template, data, writer);
	}
	

	@Override
	public void evaluateInline(Writer writer, String template, Map<String, Object> data) {
		String name = "inline://" + template;
		
		Template t = useCache ? templateCache.get(name) : null;
		if (t == null) {
			try {
				t = handlebars.compileInline(template);
			} catch (Exception e) {
				throw new RuntimeException(String.format("Inline template '%s' cannot be read", template), e);
			}
			if (useCache) {
				templateCache.put(name, t);
			}
		}
		
		evaluateTemplate(t, data, writer);
	}
	
	
	protected void evaluateTemplate(Template t, Map<String, Object> data, Writer writer) {
		  Context context = Context
				    .newBuilder(data)
				    .combine(data)
				    .resolver(JavaBeanValueResolver.INSTANCE, MapValueResolver.INSTANCE)
				    .build();

		try {
			t.apply(context, writer);
		} catch (IOException e) {
			throw new RuntimeException("Cannot evaluate template", e);
		}
	}
	
	
	/**
	 * Sets the use cache.
	 *
	 * @param useCache the new use cache
	 */
	@Value("${global.caching:true}")
	public void setUseCache(boolean useCache) {
		this.useCache = useCache;
	}

	@Autowired
	public void setTemplateLoader(TemplateLoader templateLoader) {
		this.templateLoader = templateLoader;
	}	

	
	

}
